@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2019-2024. All Rights Reserved.
+%% Copyright Ericsson AB 2019-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -449,19 +449,13 @@ gen_typeinfo(Erules, Typename, Type) ->
     end.
 
 gen_encode_prim(Erules, #type{constraint=C}=D, _Value) ->
-    BitStringConstraint = get_size_constraint(C),
+    SizeConstraint = get_size_constraint(C),
     IntConstr = int_constr(C),
     Containing = containing_constraint(Erules, C),
 
-    %% MaxBitStrSize = case BitStringConstraint of
-    %%     		[] -> none;
-    %%     		{_,'MAX'} -> none;
-    %%     		{_,Max} -> Max;
-    %%     		Max when is_integer(Max) -> Max
-    %%     	    end,
     asn1ct_name:new(enumval),
     Type = case D#type.def of
-	       'OCTET STRING'    -> maybe_legacy_octet_string();
+	       'OCTET STRING'    -> maybe_legacy_octet_string(SizeConstraint);
                'UTF8String'      -> string;
 	       'ObjectDescriptor'-> string;
 	       'NumericString'   -> string;
@@ -475,9 +469,9 @@ gen_encode_prim(Erules, #type{constraint=C}=D, _Value) ->
 	       'IA5String'       -> string;
 	       'UTCTime'         -> string;
 	       'GeneralizedTime' -> string;
-               B1 = 'BIT STRING' -> maybe_legacy_bit_string(B1,BitStringConstraint);
+               B1 = 'BIT STRING' -> maybe_legacy_bit_string(B1,SizeConstraint);
                B2 = {'BIT STRING',_NNL} -> 
-                   maybe_legacy_bit_string(B2,BitStringConstraint);
+                   maybe_legacy_bit_string(B2,SizeConstraint);
                {'INTEGER',NNL} -> {'INTEGER_NNL',NNL};
                {'ENUMERATED',{NNL,Ext}} -> {'ENUMERATED_EXT',maps:from_list(NNL++Ext)};
                {'ENUMERATED',NNL} -> {'ENUMERATED',maps:from_list(NNL)};
@@ -492,12 +486,18 @@ gen_encode_prim(Erules, #type{constraint=C}=D, _Value) ->
             {container,Type,Containing}
     end.
 
-maybe_legacy_octet_string() ->
-    case asn1ct:use_legacy_types() of
-        true ->
-            legacy_octet_string;
-        false ->
-            octet_string
+maybe_legacy_octet_string(SizeConstraint) ->
+    Type = case asn1ct:use_legacy_types() of
+               true ->
+                   legacy_octet_string;
+               false ->
+                   octet_string
+           end,
+    case SizeConstraint of
+        [] ->
+            Type;
+        _ ->
+            {Type,SizeConstraint}
     end.
 
 maybe_legacy_bit_string(BitStrType,SizeConstraint) ->
@@ -518,11 +518,12 @@ maybe_legacy_bit_string(BitStrType,SizeConstraint) ->
                     {list_to_atom(lists:concat([Type,"_nnl"])),NNL}
             end,
     case SizeConstraint of
-        S when is_integer(S) ->
-            {Type1,S};
+        [] ->
+            Type1;
         _ ->
-            Type1
+            {Type1,SizeConstraint}
     end.
+
 %%===========================================================================
 %% Generate DECODING
 %%===========================================================================
