@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2022-2024. All Rights Reserved.
+%% Copyright Ericsson AB 2022-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -47,6 +47,8 @@
          handle_resumption/2,
          send_key_update/2,
          update_cipher_key/2,
+         maybe_traffic_keylog_1_3/4,
+         maybe_forget_hs_secrets/2,
          do_maybe/0]).
 
 %%--------------------------------------------------------------------
@@ -126,7 +128,7 @@ wait_cert(internal,
             tls_gen_connection:next_event(NextState, no_record, State)
     end;
 wait_cert(info, Msg, State) ->
-    tls_gen_connection:handle_info(Msg, ?STATE(wait_cert), State);
+    tls_gen_connection:gen_info(Msg, ?STATE(wait_cert), State);
 wait_cert(Type, Msg, State) ->
     ssl_gen_statem:handle_common_event(Type, Msg, ?STATE(wait_cert), State).
 
@@ -138,7 +140,7 @@ wait_cv(internal = Type, #change_cipher_spec{} = Msg,
   when Id =/= ?EMPTY_ID ->
     handle_change_cipher_spec(Type, Msg, ?STATE(wait_cv), State);
 wait_cv(info, Msg, State) ->
-    tls_gen_connection:handle_info(Msg, ?STATE(wait_cv), State);
+    tls_gen_connection:gen_info(Msg, ?STATE(wait_cv), State);
 wait_cv(Type, Msg, State) ->
     ssl_gen_statem:handle_common_event(Type, Msg, ?STATE(wait_cv), State).
 
@@ -148,10 +150,16 @@ connection(internal, #new_session_ticket{} = NewSessionTicket, State) ->
     handle_new_session_ticket(NewSessionTicket, State),
     tls_gen_connection:next_event(?STATE(connection), no_record, State);
 
-connection(internal, #key_update{} = KeyUpdate, State0) ->
+connection(internal, #key_update{} = KeyUpdate, #state{static_env = #static_env{role = Role},
+                                                       ssl_options = SslOpts,
+                                                       protocol_specific = PS} = State0) ->
     case handle_key_update(KeyUpdate, State0) of
-        {ok, State} ->
-            tls_gen_connection:next_event(?STATE(connection), no_record, State);
+        {ok, #state{connection_states = ConnectionStates} = State} ->
+            Keep = maps:get(keep_secrets, SslOpts, false),
+            N = maps:get(num_key_updates, PS, 0),
+            maybe_traffic_keylog_1_3(Keep, Role, ConnectionStates, N),
+            tls_gen_connection:next_event(?STATE(connection), no_record,
+                                          maybe_forget_hs_secrets(Role, N, Keep, State));
         {error, State, Alert} ->
             ssl_gen_statem:handle_own_alert(Alert, ?STATE(connection), State),
             tls_gen_connection:next_event(?STATE(connection), no_record, State)
@@ -240,6 +248,16 @@ handle_resumption(#state{handshake_env = HSEnv0} = State, _) ->
     HSEnv = HSEnv0#handshake_env{resumption = true},
     State#state{handshake_env = HSEnv}.
 
+maybe_forget_hs_secrets({keylog_hs, _}, #state{connection_states = CS} = State) ->
+    %% Server finishes the handshake last and  is able to immediately forget
+    %% hs secrets used for handshake alert logging.
+    Read0 = #{security_parameters := SecParams} = ssl_record:current_connection_state(CS, read),
+    Read1 = Read0#{security_parameters => SecParams#security_parameters{client_early_data_secret = undefined}},
+    Read = maps:without([client_handshake_traffic_secret, server_handshake_traffic_secret], Read1),
+    State#state{connection_states = CS#{current_read => Read}};
+maybe_forget_hs_secrets(_, State) ->
+    State.
+
 do_maybe() ->
     Ref = erlang:make_ref(),
     Ok = fun(ok) -> ok;
@@ -289,9 +307,12 @@ send_key_update(Sender, Type) ->
     KeyUpdate = tls_handshake_1_3:key_update(Type),
     tls_sender:send_post_handshake(Sender, KeyUpdate).
 
-update_cipher_key(ConnStateName, #state{connection_states = CS0} = State0) ->
+update_cipher_key(ConnStateName, #state{connection_states = CS0,
+                                        protocol_specific = PS} = State0) ->
     CS = update_cipher_key(ConnStateName, CS0),
-    State0#state{connection_states = CS};
+    N = maps:get(num_key_updates, PS, 0),
+    State0#state{connection_states = CS,
+                 protocol_specific = PS#{num_key_updates => N + 1}};
 update_cipher_key(ConnStateName, CS0) ->
     #{security_parameters := SecParams0,
       cipher_state := CipherState0} = ConnState0 = maps:get(ConnStateName, CS0),
@@ -315,6 +336,17 @@ update_cipher_key(ConnStateName, CS0) ->
                             sequence_number => 0},
     CS0#{ConnStateName => ConnState}.
 
+maybe_traffic_keylog_1_3({keylog, Fun}, Role, ConnectionStates, N) ->
+    #{security_parameters := #security_parameters{client_random = ClientRandom,
+                                                  prf_algorithm = Prf,
+                                                  application_traffic_secret = TrafficSecret}}
+        = ssl_record:current_connection_state(ConnectionStates, read),
+    KeyLog = ssl_logger:keylog_traffic_1_3(ssl_gen_statem:opposite_role(Role), ClientRandom,
+                                           Prf, TrafficSecret, N),
+    ssl_logger:keylog(KeyLog, ClientRandom, Fun);
+maybe_traffic_keylog_1_3(_,_,_,_) ->
+    ok.
+
 %%--------------------------------------------------------------------
 do_wait_cert(#certificate_1_3{} = Certificate, State0) ->
     {Ref,Maybe} = do_maybe(),
@@ -443,3 +475,13 @@ exporter(ExporterSecret, Context0, WantedLength, PRFAlgorithm) ->
     HashContext = tls_v1:transcript_hash(Context, PRFAlgorithm),
     tls_v1:hkdf_expand_label(ExporterSecret, <<"exporter">>, HashContext,
                              WantedLength, PRFAlgorithm).
+
+%% If a client has key_logging for failing handshakes we now know that
+%% it is now safe to clear them. Could be cleared earlier but as this
+%% is debugging functionality we choose not to make complicated
+%% timeout handling to save a little memory for the client.
+%% This is not a problem on the server side.
+maybe_forget_hs_secrets(client, 0, KeepSecrets, State) ->
+    maybe_forget_hs_secrets(KeepSecrets, State);
+maybe_forget_hs_secrets(_,_,_, State) ->
+    State.
